package quimufu.colourful_portals.client;

import net.fabricmc.fabric.api.client.render.fluid.v1.FluidRenderHandler;
import net.fabricmc.fabric.api.client.render.fluid.v1.FluidRenderHandlerRegistry;
import net.minecraft.block.BlockState;
import net.minecraft.client.texture.Sprite;
import net.minecraft.fluid.FluidState;
import net.minecraft.util.math.BlockPos;
import net.minecraft.util.math.Direction;
import net.minecraft.world.BlockRenderView;
import quimufu.colourful_portals.ColourfulPortalsMod;
import quimufu.colourful_portals.portal_fluid.PortalFluid;

public class CommonPortalFluidRenderer {

    private Sprite sprite = null;
    private final float[] uIu2IvIv2 = new float[4];

    public boolean render(BlockRenderView world, FluidState fluidState, BlockPos pos, BlockPos offset, VertexEater  vertexEater) {

        FluidRenderHandler handler = FluidRenderHandlerRegistry.INSTANCE.get(fluidState.getFluid());
        sprite = handler.getFluidSprites(null, null, fluidState)[0];
        //from
        float x = 0;
        float y = 0;
        float z = 0;
        //to
        float x2 = 1;
        float y2 = 1;
        float z2 = 1;

        Direction.Axis axis = fluidState.get(PortalFluid.AXIS).getAxis();
        int amount = fluidState.get(PortalFluid.AMOUNT);
        if (amount == 0)
            return false;
        final float offsetInSquishedDir = (16 - amount) / 32f;
        if (axis != null) {
            switch (axis) {
                case X -> {
                    x += offsetInSquishedDir;
                    x2 -= offsetInSquishedDir;
                }
                case Y -> {
                    y += offsetInSquishedDir;
                    y2 -= offsetInSquishedDir;
                }
                case Z -> {
                    z += offsetInSquishedDir;
                    z2 -= offsetInSquishedDir;
                }
            }

        }

        boolean ret = false;
        for (Direction direction : Direction.values()) {

            BlockPos neighbourPos = pos.offset(direction);
            BlockState neighbour = world.getBlockState(neighbourPos);
            FluidState neighbourFluidState = neighbour.getFluidState();
            if (axis == null || direction.getAxis() != axis) {
                if (neighbour.isSideSolidFullSquare(world, neighbourPos, direction.getOpposite())
                        && neighbour.isOpaque()) {
                    continue;
                }
                if (neighbourFluidState.isOf(ColourfulPortalsMod.PORTAL_FLUID)
                        && neighbourFluidState.get(PortalFluid.AMOUNT) >= amount) {
                    continue;
                }
            }
            if (axis != null && direction.getAxis() != axis) {

                int neighbourAmount;
                if (neighbourFluidState.isOf(ColourfulPortalsMod.PORTAL_FLUID)
                        && (neighbourAmount = neighbourFluidState.get(PortalFluid.AMOUNT)) < amount) {
                    //draw side as two little silvers, adapting our width to neighbour width
                    float neighbourOffsetInSquishedDir = (16 - neighbourAmount) / 32F;
                    drawSilvers(vertexEater, direction, axis, offsetInSquishedDir, neighbourOffsetInSquishedDir, offset, (amount & 1) == 1);
                    ret = true;
                    continue;
                }
                calcSquishedUVs(axis, 1.f - offsetInSquishedDir, offsetInSquishedDir, direction, (amount & 1) == 1);
            } else {
                uIu2IvIv2[0] = sprite.getFrameU(0.F);
                uIu2IvIv2[1] = sprite.getFrameU(1.F);
                uIu2IvIv2[2] = sprite.getFrameV(0.F);
                uIu2IvIv2[3] = sprite.getFrameV(1.F);

            }
            drawSide(vertexEater, x, y, z, x2, y2, z2, direction, uIu2IvIv2[0], uIu2IvIv2[2], uIu2IvIv2[1], uIu2IvIv2[3], offset);
            ret = true;

        }
        return ret;
    }

    private void calcSquishedUVs(Direction.Axis axis, float squishedTop, float squishedBottom, Direction direction, boolean odd) {
        switch (axis) {
            case X -> {
                if (odd) {
                    uIu2IvIv2[0] = sprite.getFrameU(1.F/32.F + squishedBottom);
                    uIu2IvIv2[1] = sprite.getFrameU(1.F/32.F + squishedTop);

                } else {
                    uIu2IvIv2[0] = sprite.getFrameU(0.F + squishedBottom);
                    uIu2IvIv2[1] = sprite.getFrameU(squishedTop);
                }
                uIu2IvIv2[2] = sprite.getFrameV(0.F);
                uIu2IvIv2[3] = sprite.getFrameV(1.F);
            }
            case Z -> {
                if (direction.getAxis() == Direction.Axis.Y) {
                    if (odd) {
                        uIu2IvIv2[2] = sprite.getFrameV(1.F/32.F + squishedBottom);
                        uIu2IvIv2[3] = sprite.getFrameV(1.F/32.F + squishedTop);

                    } else {
                        uIu2IvIv2[2] = sprite.getFrameV(squishedBottom);
                        uIu2IvIv2[3] = sprite.getFrameV(squishedTop);
                    }
                    uIu2IvIv2[0] = sprite.getFrameU(0.F);
                    uIu2IvIv2[1] = sprite.getFrameU(1.F);
                } else {
                    if (odd) {
                        uIu2IvIv2[0] = sprite.getFrameU(1.F/32.F + squishedBottom);
                        uIu2IvIv2[1] = sprite.getFrameU(1.F/32.F + squishedTop);

                    } else {
                        uIu2IvIv2[0] = sprite.getFrameU(squishedBottom);
                        uIu2IvIv2[1] = sprite.getFrameU(squishedTop);
                    }
                    uIu2IvIv2[2] = sprite.getFrameV(0.F);
                    uIu2IvIv2[3] = sprite.getFrameV(1.F);

                }
            }
            case Y -> {
                if (odd) {
                    uIu2IvIv2[2] = sprite.getFrameV(1.F/32.F + squishedBottom);
                    uIu2IvIv2[3] = sprite.getFrameV(1.F/32.F + squishedTop);

                } else {
                    uIu2IvIv2[2] = sprite.getFrameV(squishedBottom);
                    uIu2IvIv2[3] = sprite.getFrameV(squishedTop);
                }
                uIu2IvIv2[0] = sprite.getFrameU(0.F);
                uIu2IvIv2[1] = sprite.getFrameU(1.F);
            }
            default -> throw new IllegalStateException("Unexpected value: " + axis);
        }
    }


    private void drawSilvers(VertexEater buffers, Direction direction, Direction.Axis axis, float offsetInSquishedDir, float neighbourOffsetInSquishedDir, BlockPos offset, boolean odd) {

        float x = 0F;
        float x2 = 1F;
        float y = 0F;
        float y2 = 1F;
        float z = 0F;
        float z2 = 1F;
        switch (axis) {
            case X -> {
                x = offsetInSquishedDir;
                x2 = neighbourOffsetInSquishedDir;
            }
            case Y -> {
                y = offsetInSquishedDir;
                y2 = neighbourOffsetInSquishedDir;
            }
            case Z -> {
                z = offsetInSquishedDir;
                z2 = neighbourOffsetInSquishedDir;
            }
        }
        calcSquishedUVs(axis, offsetInSquishedDir, neighbourOffsetInSquishedDir, direction, odd);
        drawSide(buffers, x, y, z, x2, y2, z2, direction, uIu2IvIv2[0], uIu2IvIv2[2], uIu2IvIv2[1], uIu2IvIv2[3], offset);

        switch (axis) {
            case X -> {
                x = 1F - neighbourOffsetInSquishedDir;
                x2 = 1F - offsetInSquishedDir;
            }
            case Y -> {
                y = 1F - neighbourOffsetInSquishedDir;
                y2 = 1F - offsetInSquishedDir;
            }
            case Z -> {
                z = 1F - neighbourOffsetInSquishedDir;
                z2 = 1F - offsetInSquishedDir;
            }
        }
        calcSquishedUVs(axis, 1.f - neighbourOffsetInSquishedDir,1.f - offsetInSquishedDir , direction, odd);
        drawSide(buffers, x, y, z, x2, y2, z2, direction, uIu2IvIv2[0], uIu2IvIv2[2], uIu2IvIv2[1], uIu2IvIv2[3], offset);
    }

    private void drawSide(VertexEater vertexConsumer, float x, float y, float z, float x2, float y2, float z2, Direction direction, float frameU, float frameV, float frameU2, float frameV2, BlockPos offset) {
        vertexConsumer.setSprite(sprite);
        switch (direction) {
            case DOWN -> {
                vertexConsumer.eatVertex(x, y, z, frameU, frameV2);
                vertexConsumer.eatVertex( x2, y, z, frameU2, frameV2);
                vertexConsumer.eatVertex( x2, y, z2, frameU2, frameV);
                vertexConsumer.eatVertex( x, y, z2, frameU, frameV);
                vertexConsumer.drawQuad(offset, direction);
            }
            case UP -> {
                vertexConsumer.eatVertex( x, y2, z2, frameU, frameV2);
                vertexConsumer.eatVertex( x2, y2, z2, frameU2, frameV2);
                vertexConsumer.eatVertex( x2, y2, z, frameU2, frameV);
                vertexConsumer.eatVertex( x, y2, z, frameU, frameV);
                vertexConsumer.drawQuad(offset, direction);
            }
            case NORTH -> {
                vertexConsumer.eatVertex( x, y2, z, frameU, frameV);
                vertexConsumer.eatVertex( x2, y2, z, frameU2, frameV);
                vertexConsumer.eatVertex( x2, y, z, frameU2, frameV2);
                vertexConsumer.eatVertex( x, y, z, frameU, frameV2);
                vertexConsumer.drawQuad(offset, direction);
            }
            case SOUTH -> {
                vertexConsumer.eatVertex( x, y, z2, frameU, frameV);
                vertexConsumer.eatVertex( x2, y, z2, frameU2, frameV);
                vertexConsumer.eatVertex( x2, y2, z2, frameU2, frameV2);
                vertexConsumer.eatVertex( x, y2, z2, frameU, frameV2);
                vertexConsumer.drawQuad(offset, direction);
            }
            case WEST -> {
                vertexConsumer.eatVertex( x, y, z, frameU, frameV);
                vertexConsumer.eatVertex( x, y, z2, frameU2, frameV);
                vertexConsumer.eatVertex( x, y2, z2, frameU2, frameV2);
                vertexConsumer.eatVertex( x, y2, z, frameU, frameV2);
                vertexConsumer.drawQuad(offset, direction);
            }
            case EAST -> {
                vertexConsumer.eatVertex( x2, y2, z, frameU, frameV);
                vertexConsumer.eatVertex( x2, y2, z2, frameU2, frameV);
                vertexConsumer.eatVertex( x2, y, z2, frameU2, frameV2);
                vertexConsumer.eatVertex( x2, y, z, frameU, frameV2);
                vertexConsumer.drawQuad(offset, direction);
            }
        }
    }

    public interface VertexEater {
        
        void setSprite(Sprite sprite);
        void eatVertex(float x, float y, float z, float frameU, float frameV);
        void drawQuad(BlockPos offset, Direction direction);
    }
}
