package quimufu.colourful_portals.general_util;

import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.util.*;

import static quimufu.colourful_portals.ColourfulPortalsMod.LOGGER;

public class LinkedList<E> implements List<E> {

    Node<E> headNode = null;


    @Override
    public boolean contains(Object o) {
        return indexOf(o) != -1;
    }


    @Override
    public int size() {
        Node<E> node = headNode;
        int size = 0;
        while (node != null) {
            size++;
            node = node.getNext();
        }
        LOGGER.info("" + size);
        return size;
    }

    @Override
    public boolean isEmpty() {
        return headNode == null;
    }

    @NotNull
    @Override
    public Iterator<E> iterator() {
        return new LLIterator<>(this);
    }

    @NotNull
    @Override
    public Object[] toArray() {
        Object[] objects = new Object[size()];
        Node<E> node = headNode;
        int index = 0;
        while (node != null) {
            objects[index] = node.getValue();
            node = node.getNext();
            index++;
        }

        return objects;
    }

    @NotNull
    @Override
    public <T> T @NotNull [] toArray(T[] a) {
        int size = this.size();
        if (size > a.length) {
            a = (T[])java.lang.reflect.Array.newInstance(a.getClass().getComponentType(), size);
        }
        Object[] result = a;
        Node<E> node = headNode;
        int index = 0;
        while (node != null) {
            result[index] = node.getValue();
            node = node.getNext();
            index++;
        }
        if (size != a.length) {
            a[index] = null;
        }
        return a;
    }

    @Override
    public boolean add(E e) {
        if (headNode == null) {
            headNode = new Node<>(e, this);
            return true;
        }
        Node<E> node = getLastNode(headNode);
        node.insertAfter(e);
        return true;
    }

    private @NotNull Node<E> getLastNode(@NotNull Node<E> start) {
        Node<E> node = start;
        while (node.getNext() != null) {
            node = node.getNext();
        }
        return node;
    }

    @Override
    public boolean containsAll(@NotNull Collection<?> c) {
        for (Object o : c) {
            if (!contains(o)) {
                return false;
            }
        }
        return true;
    }

    @Override
    public boolean addAll(@NotNull Collection<? extends E> c) {
        if (headNode == null) {
            headNode = new Node<>(null, this);
            Node<E> startNode = new Node<>(null, this);
            Node<E> currNode = startNode;
            for (E o : c) {
                currNode.insertAfter(o);
                currNode = currNode.getNext();
            }
            startNode.remove();
            return true;
        }
        Node<E> currNode = getLastNode(headNode);
        for (E o : c) {
            currNode.insertAfter(o);
            currNode = currNode.getNext();
        }
        return true;
    }

    @Override
    public boolean addAll(int index, @NotNull Collection<? extends E> c) {
        if (c.isEmpty()) {
            return false;
        }
        Node<E> node = getNode(index);
        for (E e : c) {
            node.insertBefore(e);
        }
        return true;
    }

    @Override
    public boolean removeAll(@NotNull Collection<?> c) {
        if (headNode == null) {
            return false;
        }
        boolean ret = false;
        Node<E> currNode = headNode;
        while (currNode != null) {
            if (c.contains(currNode.getValue())) {
                currNode.remove();
                ret = true;
            }
            currNode = currNode.getNext();
        }
        return ret;
    }

    @Override
    public boolean retainAll(@NotNull Collection<?> c) {
        if (headNode == null) {
            return false;
        }
        boolean ret = false;
        Node<E> currNode = headNode;
        while (currNode != null) {
            if (!c.contains(currNode.getValue())) {
                currNode.remove();
                ret = true;
            }
            currNode = currNode.getNext();
        }
        return ret;
    }

    @Override
    public void clear() {
        throw new UnsupportedOperationException();
    }

    @Override
    public boolean remove(Object o) {
        Node<E> node = headNode;
        while (node != null) {
            if (Objects.equals(node.getValue(), o)) {
                node.remove();
                return true;
            }
            node = node.getNext();
        }
        return false;
    }


    public E get(int index) {
        return getNode(index).getValue();
    }

    public Node<E> getNode(int index) {
        Node<E> node = headNode;
        for (int i = 0; i < index; i++) {
            if (node == null)
                throw new IndexOutOfBoundsException(index);
            node = node.getNext();

        }
        return node;
    }

    public E set(int index, E element) {
        getNode(index).setValue(element);
        return element;
    }

    public void add(int index, E element) {
        getNode(index).insertBefore(element);
    }

    public E remove(int index) {
        Node<E> node = getNode(index);
        node.remove();
        return node.getValue();
    }

    public int indexOf(Object o) {
        int index = 0;
        Node<E> node = headNode;
        while (node != null) {
            if (Objects.equals(node.getValue(), o)) {
                return index;
            }
            node = node.getNext();
            index++;
        }

        return -1;
    }

    public int lastIndexOf(Object o) {
        int ret = -1;
        int index = 0;
        Node<E> node = headNode;
        while (node != null) {
            if (Objects.equals(node.getValue(), o)) {
                ret = index;
            }
            node = node.getNext();
            index++;
        }

        return ret;
    }

    @NotNull
    @Override
    public ListIterator<E> listIterator() {

        return new LLLIterator<>(this);
    }

    @NotNull
    @Override
    public ListIterator<E> listIterator(int index) {
        LLLIterator<E> iterator = new LLLIterator<>(this);
        iterator.nextNode = getNode(index);
        iterator.previousNode = iterator.nextNode.getPrevious();
        return iterator;
    }

    @NotNull
    @Override
    public List<E> subList(int fromIndex, int toIndex) {
        throw new UnsupportedOperationException();
    }

    @Nullable
    public Node<E> getNodeOf(E content) {
        Node<E> node = headNode;
        while (node != null) {
            if (Objects.equals(node.getValue(), content)) {
                return node;
            }
            node = node.getNext();
        }

        return null;
    }

    private static class LLIterator<E> implements Iterator<E> {

        private Node<E> nextNode;

        public LLIterator(LinkedList<E> linkedList) {
            nextNode = linkedList.headNode;
        }

        @Override
        public boolean hasNext() {
            return nextNode != null;
        }

        @Override
        public E next() {
            if (nextNode == null)
                throw new IndexOutOfBoundsException();

            E value = nextNode.getValue();
            nextNode = nextNode.getNext();
            return value;
        }
    }

    private static class LLLIterator<E> implements ListIterator<E> {

        @Nullable
        private Node<E> nextNode;
        @Nullable
        private Node<E> previousNode;
        private boolean lastNext = true;

        private final LinkedList<E> linkedList;

        public LLLIterator(LinkedList<E> linkedList) {
            nextNode = linkedList.headNode;
            previousNode = null;
            this.linkedList = linkedList;
        }

        @Override
        public boolean hasNext() {
            return nextNode != null;
        }

        @Override
        public E next() {
            tryFixOrphans();
            if (nextNode == null)
                throw new IndexOutOfBoundsException();
            
            E value = nextNode.getValue();
            previousNode = nextNode;
            nextNode = nextNode.getNext();
            lastNext = true;
            return value;
        }
        
        @Override
        public boolean hasPrevious() {
            return previousNode != null;
        }

        @Override
        public E previous() {
            tryFixOrphans();
            if (previousNode == null)
                throw new IndexOutOfBoundsException();
            E value = previousNode.getValue();
            nextNode = previousNode;
            previousNode = previousNode.getPrevious();
            lastNext = false;
            return value;
        }

        @Override
        public int nextIndex() {
            tryFixOrphans();
            if(nextNode == null)
                return linkedList.size();
            return linkedList.indexOfNode(nextNode);
        }

        @Override
        public int previousIndex() {
            tryFixOrphans();
            if(previousNode == null) {
                return -1;
            }
            return linkedList.indexOfNode(previousNode);
        }

        @Override
        public void remove() {
            tryFixOrphans();
            if(lastNext) {
                if(previousNode == null)
                    throw new IllegalStateException();
                previousNode.remove();
                previousNode = previousNode.getPrevious();
            } else {
                if(nextNode == null)
                    throw new IllegalStateException();
                nextNode.remove();
                nextNode = nextNode.getNext();
                
            }
        }

        @Override
        public void set(E e) {
            tryFixOrphans();
            if(lastNext) {
                if(previousNode == null)
                    throw new IllegalStateException();
                previousNode.setValue(e);
            } else {
                if(nextNode == null)
                    throw new IllegalStateException();
                nextNode.setValue(e);
            }
        }

        @Override
        public void add(E e) {
            tryFixOrphans();
            if(nextNode != null) {
                nextNode.insertBefore(e);
            } else if(previousNode != null) {
                previousNode.insertAfter(e);
            } else {
                // empty!
                linkedList.add(e);
                nextNode = linkedList.headNode;
            }

        }

        private void tryFixOrphans() {
            if (previousNode != null && previousNode.orphaned()){
                if(nextNode != null && !nextNode.orphaned()){
                    previousNode = nextNode.getPrevious();
                } else {
                    throw new ConcurrentModificationException("Node was concurrently deleted!");
                }
            }
            if (nextNode != null && nextNode.orphaned()){
                if(previousNode != null && !previousNode.orphaned()){
                    nextNode = previousNode.getNext();
                } else {
                    throw new ConcurrentModificationException("Node was concurrently deleted!");
                }
            }
        }
    }

    private int indexOfNode(Node<E> needle) {
        int index = 0;
        Node<E> node = headNode;
        while (node != null) {
            if (node == needle) {
                return index;
            }
            node = node.getNext();
            index++;
        }

        return -1;
    }
}
