package quimufu.colourful_portals.portal;

import net.minecraft.nbt.NbtCompound;
import net.minecraft.nbt.NbtElement;
import net.minecraft.nbt.NbtHelper;
import net.minecraft.nbt.NbtList;
import net.minecraft.registry.RegistryWrapper;
import net.minecraft.util.Identifier;
import net.minecraft.util.math.BlockBox;
import net.minecraft.util.math.BlockPos;
import net.minecraft.world.WorldProperties;
import org.ladysnake.cca.api.v3.component.Component;
import quimufu.colourful_portals.general_util.LinkedList;
import quimufu.colourful_portals.general_util.Node;

import java.util.HashMap;
import java.util.List;
import java.util.Set;

import static quimufu.colourful_portals.ColourfulPortalsMod.*;

public class PortalListComponent implements Component {
    public static final int CURRENT_VERSION = 1;
    private final WorldProperties worldProperties;
    HashMap<Identifier, LinkedList<PortalRepresentation>> portalsPerPortalBlock = new HashMap<>();

    Identifier lastPortalLinkingSystem = Identifier.of(MOD_ID, "immersive_portals_linking_system");

    public PortalListComponent(WorldProperties worldProperties) {
        this.worldProperties = worldProperties;
    }

    @Override
    public void readFromNbt(NbtCompound tag, RegistryWrapper.WrapperLookup registryLookup) {
        if (tag != null && !tag.isEmpty()) {
            if (!tag.contains("version")) {
                readFromOldNbt(tag);
            }
            if (tag.getInt("version") > CURRENT_VERSION) {
                LOGGER.warn("portals component comes from a newer version! Hopefully backwards compatible. Proceeding");
            }
            if (tag.contains("linking_system")) {
                lastPortalLinkingSystem = Identifier.tryParse(tag.getString("version"));
            }
            NbtCompound blocks = tag.getCompound("blocks");
            for (String block : blocks.getKeys()) {
                NbtList portalsWithDim = blocks.getList(block, NbtElement.COMPOUND_TYPE);
                LinkedList<PortalRepresentation> portalRepresentations = new LinkedList<>();
                for (int i = 0; i < portalsWithDim.size(); i++) {
                    NbtCompound portalWithDim = portalsWithDim.getCompound(i);

                    NbtCompound portalCompound = portalWithDim.getCompound("portal");
                    BlockBox portal = BlockBox.create(NbtHelper.toBlockPos(portalCompound, "from").orElseThrow(),
                            NbtHelper.toBlockPos(portalCompound, "to").orElseThrow());

                    Identifier dimension = Identifier.tryParse(portalWithDim.getString("dim"));

                    portalRepresentations.add(new PortalRepresentation(portal, dimension));
                }
                portalsPerPortalBlock.put(Identifier.tryParse(block), portalRepresentations);

            }
        }
    }

    private void readFromOldNbt(NbtCompound tag) {
        for (String block : tag.getKeys()) {
            NbtList portalsWithDim = tag.getList(block, NbtElement.COMPOUND_TYPE);
            LinkedList<PortalRepresentation> portalRepresentations = new LinkedList<>();
            for (int i = 0; i < portalsWithDim.size(); i++) {
                NbtCompound portalWithDim = portalsWithDim.getCompound(i);

                NbtCompound portalCompound = portalWithDim.getCompound("portal");
                BlockBox portal = BlockBox.create(toBlockPosOld(portalCompound.getCompound("from")),
                        toBlockPosOld(portalCompound.getCompound("to")));

                Identifier dimension = Identifier.tryParse(portalWithDim.getString("dim"));

                portalRepresentations.add(new PortalRepresentation(portal, dimension));
            }
            portalsPerPortalBlock.put(Identifier.tryParse(block), portalRepresentations);
        }
    }

    public static BlockPos toBlockPosOld(NbtCompound nbt) {
        return new BlockPos(nbt.getInt("X"), nbt.getInt("Y"), nbt.getInt("Z"));
    }

    @Override
    public void writeToNbt(NbtCompound tag, RegistryWrapper.WrapperLookup registryLookup) {
        LOGGER.debug("save");
        if(PORTAL_MANAGER == null) {
            return;
        }
        tag.putInt("version", CURRENT_VERSION);
        tag.putString("linking_system", PORTAL_MANAGER.getLinkingSystem().getLinkingSystemId().toString());
        NbtCompound blocks = new NbtCompound();
        tag.put("blocks", blocks);
        for (Identifier portalBlockId : portalsPerPortalBlock.keySet()) {
            NbtList portalsWithDimList = new NbtList();
            blocks.put(portalBlockId.toString(), portalsWithDimList);
            for (PortalRepresentation portalRepresentationWithDim : portalsPerPortalBlock.get(portalBlockId)) {
                NbtCompound portalWithDimCompound = new NbtCompound();
                portalsWithDimList.add(portalWithDimCompound);

                NbtCompound portalCompound = new NbtCompound();
                portalWithDimCompound.put("portal", portalCompound);
                portalWithDimCompound.putString("dim", portalRepresentationWithDim.dimensionId().toString());

                BlockBox portal = portalRepresentationWithDim.location();
                BlockPos from = new BlockPos(portal.getMinX(), portal.getMinY(), portal.getMinZ());
                BlockPos to = new BlockPos(portal.getMaxX(), portal.getMaxY(), portal.getMaxZ());

                portalCompound.put("from", NbtHelper.fromBlockPos(from));
                portalCompound.put("to", NbtHelper.fromBlockPos(to));
            }

        }
        LOGGER.info("portals {}", tag);


    }

    public List<PortalRepresentation> getContainingPortals(Identifier blockId, BlockPos pos, Identifier dim) {
        return getPortals(blockId).stream()
                .filter((portal) -> (portal.dimensionId().equals(dim) && portal.location().contains(pos)))
                .toList();
    }

    public void createPortal(Identifier blockId, PortalRepresentation portal) {
        LinkedList<PortalRepresentation> portalRepresentations = getPortals(blockId);
        if (!portalRepresentations.contains(portal)) {
            portalRepresentations.add(portal);
        }
    }

    public void removePortal(Identifier blockId, PortalRepresentation portal) {
        getPortals(blockId).remove(portal);
    }

    public LinkedList<PortalRepresentation> getPortals(Identifier blockId) {
        return portalsPerPortalBlock.computeIfAbsent(blockId, (i) -> new LinkedList<>());
    }

    public Set<Identifier> getBlockIds() {
        return portalsPerPortalBlock.keySet();
    }

    public PortalRepresentation getNext(Identifier blockId, PortalRepresentation portalRepresentationWithDim) {
        LOGGER.info("gn");
        LinkedList<PortalRepresentation> portals = getPortals(blockId);
        Node<PortalRepresentation> node;
        if ((node = portals.getNodeOf(portalRepresentationWithDim)) == null) {

            return portals.getFirst();
        }
        PortalRepresentation portalRepresentation = node.getNext() == null ? portals.getFirst() : node.getNext().getValue();
        LOGGER.info("gne");
        return portalRepresentation;
    }

    public boolean containsPortal(Identifier blockId, PortalRepresentation portalRepresentation) {
        return getPortals(blockId).contains(portalRepresentation);
    }

    public PortalRepresentation getLast(Identifier blockId) {
        return getPortals(blockId).getLast();
    }

    public Identifier lastPortalLinkingSystemId() {
        return lastPortalLinkingSystem;
    }
}
