package quimufu.colourful_portals.portal_fluid;

import net.minecraft.block.Block;
import net.minecraft.block.BlockState;
import net.minecraft.block.FluidFillable;
import net.minecraft.entity.player.PlayerEntity;
import net.minecraft.item.BucketItem;
import net.minecraft.item.ItemStack;
import net.minecraft.sound.SoundCategory;
import net.minecraft.sound.SoundEvent;
import net.minecraft.sound.SoundEvents;
import net.minecraft.util.Hand;
import net.minecraft.util.TypedActionResult;
import net.minecraft.util.hit.BlockHitResult;
import net.minecraft.util.hit.HitResult;
import net.minecraft.util.math.BlockPos;
import net.minecraft.world.RaycastContext;
import net.minecraft.world.World;
import net.minecraft.world.WorldAccess;
import net.minecraft.world.event.GameEvent;

public class PortalFluidBucketItem extends BucketItem {
    private final PortalFluid fluid;

    public PortalFluidBucketItem(PortalFluid fluid, Settings settings) {
        super(fluid, settings);
        this.fluid = fluid;
    }

    @Override
    public TypedActionResult<ItemStack> use(World world, PlayerEntity user, Hand hand) {
        BlockHitResult blockHitResult = BucketItem.raycast(world, user, RaycastContext.FluidHandling.NONE);
        ItemStack itemStack = user.getStackInHand(hand);
        if (blockHitResult.getType() == HitResult.Type.BLOCK) {
            BlockPos pos = blockHitResult.getBlockPos();
            BlockState blockState = world.getBlockState(pos);
            Block block = blockState.getBlock();
            if (block instanceof FluidFillable && ((FluidFillable) block).canFillWithFluid(user, world, pos, blockState, fluid)) {
                if (placeFluid(user, world, pos, blockHitResult)) {
                    return TypedActionResult.success(BucketItem.getEmptiedStack(itemStack, user), world.isClient());
                }
            }
        }

        blockHitResult = BucketItem.raycast(world, user, RaycastContext.FluidHandling.ANY);
        itemStack = user.getStackInHand(hand);
        if (blockHitResult.getType() == HitResult.Type.BLOCK) {
            BlockPos pos = blockHitResult.getBlockPos();
            BlockState blockState = world.getBlockState(pos);
            if (blockState.getFluidState().isOf(fluid)) {
                if (placeFluid(user, world, pos, blockHitResult)) {
                    return TypedActionResult.success(BucketItem.getEmptiedStack(itemStack, user), world.isClient());
                }
            }
        }

        return super.use(world, user, hand);
    }

    @Override
    public boolean placeFluid(PlayerEntity player, World world, BlockPos pos, BlockHitResult hitResult) {
        BlockState blockState = world.getBlockState(pos);
        Block block = blockState.getBlock();
        boolean shouldTryPlace = blockState.isAir()
                || blockState.canBucketPlace(this.fluid)
                || block instanceof FluidFillable && ((FluidFillable) block).canFillWithFluid(player, world, pos, blockState, this.fluid);
        if (!shouldTryPlace) {
            return hitResult != null && this.placeFluid(player, world, hitResult.getBlockPos().offset(hitResult.getSide()), null);
        }
        if (blockState.getFluidState().isOf(fluid)) {
            if (fluid.tryFillWithFluid(world, pos, blockState, this.fluid.getDefaultState())) {
                this.playEmptyingSound(player, world, pos);
                return true;
            }
            return false;
        }
        if (block instanceof FluidFillable) {
            ((FluidFillable) block).tryFillWithFluid(world, pos, blockState, this.fluid.getDefaultState());
            this.playEmptyingSound(player, world, pos);
            return true;
        }


        if (!world.isClient && blockState.canBucketPlace(this.fluid) && !blockState.isLiquid()) {
            world.breakBlock(pos, true);
        }
        if (world.setBlockState(pos, this.fluid.getDefaultState().getBlockState(), Block.NOTIFY_ALL | Block.REDRAW_ON_MAIN_THREAD) || blockState.getFluidState().isStill()) {
            this.playEmptyingSound(player, world, pos);
            return true;
        }
        return false;
    }

    //we need to overwrite this because the fabric API mixin crashes otherwise!
    @Override
    protected void playEmptyingSound(PlayerEntity player, WorldAccess world, BlockPos pos) {
        SoundEvent soundEvent = SoundEvents.ITEM_BUCKET_EMPTY;
        world.playSound(player, pos, soundEvent, SoundCategory.BLOCKS, 1.0f, 1.0f);
        world.emitGameEvent(player, GameEvent.FLUID_PLACE, pos);
    }
}
